"""
AMRVAC data structures



"""

import os
import struct
import warnings
import weakref
from pathlib import Path

import numpy as np
from more_itertools import always_iterable

from yt.config import ytcfg
from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.static_output import Dataset
from yt.funcs import mylog, setdefaultattr
from yt.geometry.api import Geometry
from yt.geometry.grid_geometry_handler import GridIndex
from yt.utilities.physical_constants import boltzmann_constant_cgs as kb_cgs

from .datfile_utils import get_header, get_tree_info
from .fields import AMRVACFieldInfo
from .io import read_amrvac_namelist


def _parse_geometry(geometry_tag: str) -> Geometry:
    """Translate AMRVAC's geometry tag to yt's format.

    Parameters
    ----------
    geometry_tag : str
        A geometry tag as read from AMRVAC's datfile from v5.

    Returns
    -------
    geometry_yt : Geometry
        An enum member of the yt.geometry.geometry_enum.Geometry class

    Examples
    --------
    >>> _parse_geometry("Polar_2.5D")
     <Geometry.POLAR: 'polar'>
    >>> _parse_geometry("Cartesian_2.5D")
    <Geometry.CARTESIAN: 'cartesian'>

    """
    geometry_str, _, _dimension_str = geometry_tag.partition("_")
    return Geometry(geometry_str.lower())


class AMRVACGrid(AMRGridPatch):
    """A class to populate AMRVACHierarchy.grids, setting parent/children relations."""

    _id_offset = 0

    def __init__(self, id, index, level):
        # <level> should use yt's convention (start from 0)
        super().__init__(id, filename=index.index_filename, index=index)
        self.Parent = None
        self.Children = []
        self.Level = level

    def get_global_startindex(self):
        """Refresh and retrieve the starting index for each dimension at current level.

        Returns
        -------
        self.start_index : int
        """
        start_index = (self.LeftEdge - self.ds.domain_left_edge) / self.dds
        self.start_index = np.rint(start_index).astype("int64").ravel()
        return self.start_index

    def retrieve_ghost_zones(self, n_zones, fields, all_levels=False, smoothed=False):
        if smoothed:
            warnings.warn(
                "ghost-zones interpolation/smoothing is not "
                "currently supported for AMRVAC data.",
                category=RuntimeWarning,
                stacklevel=2,
            )
            smoothed = False
        return super().retrieve_ghost_zones(
            n_zones, fields, all_levels=all_levels, smoothed=smoothed
        )


class AMRVACHierarchy(GridIndex):
    grid = AMRVACGrid

    def __init__(self, ds, dataset_type="amrvac"):
        self.dataset_type = dataset_type
        self.dataset = weakref.proxy(ds)
        # the index file *is* the datfile
        self.index_filename = self.dataset.parameter_filename
        self.directory = os.path.dirname(self.index_filename)
        self.float_type = np.float64

        super().__init__(ds, dataset_type)

    def _detect_output_fields(self):
        """Parse field names from the header, as stored in self.dataset.parameters"""
        self.field_list = [
            (self.dataset_type, f) for f in self.dataset.parameters["w_names"]
        ]

    def _count_grids(self):
        """Set self.num_grids from datfile header."""
        self.num_grids = self.dataset.parameters["nleafs"]

    def _parse_index(self):
        """Populate self.grid_* attributes from tree info from datfile header."""
        with open(self.index_filename, "rb") as istream:
            vaclevels, morton_indices, block_offsets = get_tree_info(istream)
            assert (
                len(vaclevels)
                == len(morton_indices)
                == len(block_offsets)
                == self.num_grids
            )

        self.block_offsets = block_offsets
        # YT uses 0-based grid indexing:
        # lowest level = 0, while AMRVAC uses 1 for lowest level
        ytlevels = np.array(vaclevels, dtype="int32") - 1
        self.grid_levels.flat[:] = ytlevels
        self.min_level = np.min(ytlevels)
        self.max_level = np.max(ytlevels)
        assert self.max_level == self.dataset.parameters["levmax"] - 1

        # some aliases for left/right edges computation in the coming loop
        domain_width = self.dataset.parameters["xmax"] - self.dataset.parameters["xmin"]
        block_nx = self.dataset.parameters["block_nx"]
        xmin = self.dataset.parameters["xmin"]
        dx0 = (
            domain_width / self.dataset.parameters["domain_nx"]
        )  # dx at coarsest grid level (YT level 0)
        dim = self.dataset.dimensionality

        self.grids = np.empty(self.num_grids, dtype="object")
        for igrid, (ytlevel, morton_index) in enumerate(
            zip(ytlevels, morton_indices, strict=True)
        ):
            dx = dx0 / self.dataset.refine_by**ytlevel
            left_edge = xmin + (morton_index - 1) * block_nx * dx

            # edges and dimensions are filled in a dimensionality-agnostic way
            self.grid_left_edge[igrid, :dim] = left_edge
            self.grid_right_edge[igrid, :dim] = left_edge + block_nx * dx
            self.grid_dimensions[igrid, :dim] = block_nx
            self.grids[igrid] = self.grid(igrid, self, ytlevels[igrid])

    def _populate_grid_objects(self):
        # required method
        for g in self.grids:
            g._prepare_grid()
            g._setup_dx()


class AMRVACDataset(Dataset):
    _index_class = AMRVACHierarchy
    _field_info_class = AMRVACFieldInfo

    def __init__(
        self,
        filename,
        dataset_type="amrvac",
        units_override=None,
        unit_system="cgs",
        geometry_override=None,
        parfiles=None,
        default_species_fields=None,
    ):
        """Instantiate AMRVACDataset.

        Parameters
        ----------
        filename : str
            Path to a datfile.

        dataset_type : str, optional
            This should always be 'amrvac'.

        units_override : dict, optional
            A dictionary of physical normalisation factors to interpret on disk data.

        unit_system : str, optional
            Either "cgs" (default), "mks" or "code"

        geometry_override : str, optional
            A geometry flag formatted either according to either
            AMRVAC or yt standards.
            When this parameter is passed along with v5 or more newer datfiles,
            will precede over their internal "geometry" tag.

        parfiles : str or list, optional
            One or more parfiles to be passed to
            yt.frontends.amrvac.read_amrvac_parfiles()

        """
        # note: geometry_override and parfiles are specific to this frontend

        self._geometry_override = geometry_override
        self._parfiles = []

        if parfiles is not None:
            parfiles = list(always_iterable(parfiles))
            ppf = Path(parfiles[0])
            if not ppf.is_absolute() and Path(filename).resolve().is_relative_to(
                ytcfg["yt", "test_data_dir"]
            ):
                mylog.debug(
                    "Looks like %s is relative to your test_data_dir. Assuming this is also true for parfiles.",
                    filename,
                )
                parfiles = [Path(ytcfg["yt", "test_data_dir"]) / pf for pf in parfiles]
            self._parfiles = parfiles

        super().__init__(
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )

        namelist = None
        namelist_gamma = None
        c_adiab = None
        e_is_internal = None

        if parfiles is not None:
            namelist = read_amrvac_namelist(parfiles)
            if "hd_list" in namelist:
                c_adiab = namelist["hd_list"].get("hd_adiab", 1.0)
                namelist_gamma = namelist["hd_list"].get("hd_gamma")
            elif "mhd_list" in namelist:
                c_adiab = namelist["mhd_list"].get("mhd_adiab", 1.0)
                namelist_gamma = namelist["mhd_list"].get("mhd_gamma")

            if namelist_gamma is not None and self.gamma != namelist_gamma:
                mylog.error(
                    "Inconsistent values in gamma: datfile %s, parfiles %s",
                    self.gamma,
                    namelist_gamma,
                )

            if "method_list" in namelist:
                e_is_internal = namelist["method_list"].get("solve_internal_e", False)

        if c_adiab is not None:
            # this complicated unit is required for the adiabatic equation
            # of state to make physical sense
            c_adiab *= (
                self.mass_unit ** (1 - self.gamma)
                * self.length_unit ** (2 + 3 * (self.gamma - 1))
                / self.time_unit**2
            )

        self.namelist = namelist
        self._c_adiab = c_adiab
        self._e_is_internal = e_is_internal

        self.fluid_types += ("amrvac",)
        # refinement factor between a grid and its subgrid
        self.refine_by = 2

    @classmethod
    def _is_valid(cls, filename: str, *args, **kwargs) -> bool:
        """At load time, check whether data is recognized as AMRVAC formatted."""
        validation = False
        if filename.endswith(".dat"):
            try:
                with open(filename, mode="rb") as istream:
                    fmt = "=i"
                    [datfile_version] = struct.unpack(
                        fmt, istream.read(struct.calcsize(fmt))
                    )
                    if 3 <= datfile_version < 6:
                        fmt = "=ii"
                        offset_tree, offset_blocks = struct.unpack(
                            fmt, istream.read(struct.calcsize(fmt))
                        )
                        istream.seek(0, 2)
                        file_size = istream.tell()
                        validation = (
                            offset_tree < file_size and offset_blocks < file_size
                        )
            except Exception:
                pass
        return validation

    def _parse_parameter_file(self):
        """Parse input datfile's header. Apply geometry_override if specified."""
        # required method

        # populate self.parameters with header data
        with open(self.parameter_filename, "rb") as istream:
            self.parameters.update(get_header(istream))

        self.current_time = self.parameters["time"]
        self.dimensionality = self.parameters["ndim"]

        # force 3D for this definition
        dd = np.ones(3, dtype="int64")
        dd[: self.dimensionality] = self.parameters["domain_nx"]
        self.domain_dimensions = dd

        if self.parameters.get("staggered", False):
            mylog.warning(
                "'staggered' flag was found, but is currently ignored (unsupported)"
            )

        # parse geometry
        # by order of decreasing priority, we use
        # - geometry_override
        # - "geometry" parameter from datfile
        # - if all fails, default to "cartesian"
        self.geometry = Geometry.CARTESIAN

        amrvac_geom = self.parameters.get("geometry", None)
        if amrvac_geom is not None:
            self.geometry = _parse_geometry(amrvac_geom)
        elif self.parameters["datfile_version"] > 4:
            mylog.error(
                "No 'geometry' flag found in datfile with version %d >4.",
                self.parameters["datfile_version"],
            )

        if self._geometry_override is not None:
            try:
                new_geometry = _parse_geometry(self._geometry_override)
                if new_geometry == self.geometry:
                    mylog.info("geometry_override is identical to datfile parameter.")
                else:
                    self.geometry = new_geometry
                    mylog.warning(
                        "Overriding geometry, this may lead to surprising results."
                    )
            except ValueError:
                mylog.error(
                    "Unable to parse geometry_override '%s' (will be ignored).",
                    self._geometry_override,
                )

        # parse peridiocity
        periodicity = self.parameters.get("periodic", ())
        missing_dim = 3 - len(periodicity)
        self._periodicity = (*periodicity, *(missing_dim * (False,)))

        self.gamma = self.parameters.get("gamma", 5.0 / 3.0)

        # parse domain edges
        dle = np.zeros(3)
        dre = np.ones(3)
        dle[: self.dimensionality] = self.parameters["xmin"]
        dre[: self.dimensionality] = self.parameters["xmax"]
        self.domain_left_edge = dle
        self.domain_right_edge = dre

        # defaulting to non-cosmological
        self.cosmological_simulation = 0
        self.current_redshift = 0.0
        self.omega_matter = 0.0
        self.omega_lambda = 0.0
        self.hubble_constant = 0.0

    # units stuff ======================================================================
    def _set_code_unit_attributes(self):
        """Reproduce how AMRVAC internally set up physical normalisation factors."""
        # This gets called later than Dataset._override_code_units()
        # This is the reason why it uses setdefaultattr: it will only fill in the gaps
        # left by the "override", instead of overriding them again.

        # note: yt sets hydrogen mass equal to proton mass, amrvac doesn't.
        mp_cgs = self.quan(1.672621898e-24, "g")  # This value is taken from AstroPy

        # get self.length_unit if overrides are supplied, otherwise use default
        length_unit = getattr(self, "length_unit", self.quan(1, "cm"))
        namelist = read_amrvac_namelist(self._parfiles)
        He_abundance = namelist.get("mhd_list", {}).get("he_abundance", 0.1)

        # 1. calculations for mass, density, numberdensity
        if "mass_unit" in self.units_override:
            # in this case unit_mass is supplied (and has been set as attribute)
            mass_unit = self.mass_unit
            density_unit = mass_unit / length_unit**3
            nd_unit = density_unit / ((1.0 + 4.0 * He_abundance) * mp_cgs)
        else:
            # other case: numberdensity is supplied.
            # Fall back to one (default) if no overrides supplied
            try:
                nd_unit = self.quan(self.units_override["numberdensity_unit"])
            except KeyError:
                nd_unit = self.quan(
                    1.0, self.__class__.default_units["numberdensity_unit"]
                )
            density_unit = (1.0 + 4.0 * He_abundance) * mp_cgs * nd_unit
            mass_unit = density_unit * length_unit**3

        # 2. calculations for velocity
        if "time_unit" in self.units_override:
            # in this case time was supplied
            velocity_unit = length_unit / self.time_unit
        else:
            # other case: velocity was supplied.
            # Fall back to None if no overrides supplied
            velocity_unit = getattr(self, "velocity_unit", None)

        # 3. calculations for pressure and temperature
        if velocity_unit is None:
            # velocity and time not given, see if temperature is given.
            # Fall back to one (default) if not
            temperature_unit = getattr(self, "temperature_unit", self.quan(1, "K"))
            pressure_unit = (
                (2.0 + 3.0 * He_abundance) * nd_unit * kb_cgs * temperature_unit
            ).in_cgs()
            velocity_unit = (np.sqrt(pressure_unit / density_unit)).in_cgs()
        else:
            # velocity is not zero if either time was given OR velocity was given
            pressure_unit = (density_unit * velocity_unit**2).in_cgs()
            temperature_unit = (
                pressure_unit / ((2.0 + 3.0 * He_abundance) * nd_unit * kb_cgs)
            ).in_cgs()

        # 4. calculations for magnetic unit and time
        time_unit = getattr(
            self, "time_unit", length_unit / velocity_unit
        )  # if time given use it, else calculate
        magnetic_unit = (np.sqrt(4 * np.pi * pressure_unit)).to("gauss")

        setdefaultattr(self, "mass_unit", mass_unit)
        setdefaultattr(self, "density_unit", density_unit)

        setdefaultattr(self, "length_unit", length_unit)
        setdefaultattr(self, "velocity_unit", velocity_unit)
        setdefaultattr(self, "time_unit", time_unit)

        setdefaultattr(self, "temperature_unit", temperature_unit)
        setdefaultattr(self, "pressure_unit", pressure_unit)
        setdefaultattr(self, "magnetic_unit", magnetic_unit)

    allowed_unit_combinations = [
        {"numberdensity_unit", "temperature_unit", "length_unit"},
        {"mass_unit", "temperature_unit", "length_unit"},
        {"mass_unit", "time_unit", "length_unit"},
        {"numberdensity_unit", "velocity_unit", "length_unit"},
        {"mass_unit", "velocity_unit", "length_unit"},
    ]

    default_units = {
        "length_unit": "cm",
        "time_unit": "s",
        "mass_unit": "g",
        "velocity_unit": "cm/s",
        "magnetic_unit": "gauss",
        "temperature_unit": "K",
        # this is the one difference with Dataset.default_units:
        # we accept numberdensity_unit as a valid override
        "numberdensity_unit": "cm**-3",
    }

    @classmethod
    def _validate_units_override_keys(cls, units_override):
        """Check that keys in units_override are consistent with AMRVAC's internal
        normalisations factors.
        """
        # YT supports overriding other normalisations, this method ensures consistency
        # between supplied 'units_override' items and those used by AMRVAC.

        # AMRVAC's normalisations/units have 3 degrees of freedom.
        # Moreover, if temperature unit is specified then velocity unit will be
        # calculated accordingly, and vice-versa.
        # We replicate this by allowing a finite set of combinations.

        # there are only three degrees of freedom, so explicitly check for this
        if len(units_override) > 3:
            raise ValueError(
                "More than 3 degrees of freedom were specified "
                f"in units_override ({len(units_override)} given)"
            )
        # temperature and velocity cannot both be specified
        if "temperature_unit" in units_override and "velocity_unit" in units_override:
            raise ValueError(
                "Either temperature or velocity is allowed in units_override, not both."
            )
        # check if provided overrides are allowed
        suo = set(units_override)
        for allowed_combo in cls.allowed_unit_combinations:
            if suo.issubset(allowed_combo):
                break
        else:
            raise ValueError(
                f"Combination {suo} passed to units_override "
                "is not consistent with AMRVAC.\n"
                f"Allowed combinations are {cls.allowed_unit_combinations}"
            )

        # syntax for mixing super with classmethod is weird...
        super(cls, cls)._validate_units_override_keys(units_override)



from collections import OrderedDict

from yt.data_objects.particle_filters import add_particle_filter
from yt.frontends.owls.api import OWLSDataset
from yt.testing import ParticleSelectionComparison, requires_file, requires_module
from yt.utilities.answer_testing.framework import data_dir_load, requires_ds, sph_answer

os33 = "snapshot_033/snap_033.0.hdf5"

# This maps from field names to weight field names to use for projections
_fields = OrderedDict(
    [
        (("gas", "density"), None),
        (("gas", "temperature"), None),
        (("gas", "temperature"), ("gas", "density")),
        (("gas", "He_p0_number_density"), None),
        (("gas", "velocity_magnitude"), None),
    ]
)


@requires_module("h5py")
@requires_ds(os33, big_data=True)
def test_snapshot_033():
    ds = data_dir_load(os33)
    psc = ParticleSelectionComparison(ds)
    psc.run_defaults()
    for test in sph_answer(ds, "snap_033", 2 * 128**3, _fields):
        test_snapshot_033.__name__ = test.description
        yield test


@requires_module("h5py")
@requires_file(os33)
def test_OWLSDataset():
    assert isinstance(data_dir_load(os33), OWLSDataset)


@requires_module("h5py")
@requires_ds(os33)
def test_OWLS_particlefilter():
    ds = data_dir_load(os33)
    ad = ds.all_data()

    def cold_gas(pfilter, data):
        temperature = data[pfilter.filtered_type, "Temperature"]
        filter = temperature.in_units("K") <= 1e5
        return filter

    add_particle_filter(
        "gas_cold",
        function=cold_gas,
        filtered_type="PartType0",
        requires=["Temperature"],
    )
    ds.add_particle_filter("gas_cold")

    mask = ad["PartType0", "Temperature"] <= 1e5
    assert (
        ad["PartType0", "Temperature"][mask].shape
        == ad["gas_cold", "Temperature"].shape
    )



cimport numpy as np
from libc.stdint cimport int32_t, int64_t, uint32_t, uint64_t
from libcpp cimport bool
from libcpp.pair cimport pair
from libcpp.vector cimport vector


cdef extern from "<iostream>" namespace "std":
    cdef cppclass istream:
        pass

    cdef cppclass ostream:
        pass

# the following extern definitions adapted from
# http://stackoverflow.com/a/31009461/1382869

# obviously std::ios_base isn't a namespace, but this lets
# Cython generate the correct C++ code
cdef extern from "<iostream>" namespace "std::ios_base":
    cdef cppclass open_mode:
        pass
    cdef open_mode binary
    # you can define other constants as needed

cdef extern from "<fstream>" namespace "std":
    cdef cppclass ofstream(ostream):
        # constructors
        ofstream(const char*) except +
        ofstream(const char*, open_mode) except+

    cdef cppclass ifstream(istream):
        # constructors
        ifstream(const char*) except +
        ifstream(const char*, open_mode) except+


cdef extern from "c_kdtree.hpp":
    cdef cppclass Node:
        bool is_leaf
        uint32_t leafid
        uint32_t ndim
        double *left_edge
        double *right_edge
        uint64_t left_idx
        uint64_t children
        uint32_t split_dim
        double split
        Node* less
        Node* greater
        bool *periodic_left
        bool *periodic_right
        vector[vector[uint32_t]] left_neighbors
        vector[vector[uint32_t]] right_neighbors
        vector[uint32_t] all_neighbors
    cdef cppclass KDTree:
        uint64_t* all_idx
        uint64_t npts
        uint32_t ndim
        int64_t data_version
        uint32_t leafsize
        double* domain_left_edge
        double* domain_right_edge
        double* domain_width
        double* domain_mins
        double* domain_maxs
        bool* periodic
        bool any_periodic
        uint32_t num_leaves
        vector[Node*] leaves
        Node* root
        KDTree(double *pts, uint64_t *idx, uint64_t n, uint32_t m,
               uint32_t leafsize0, double *left_edge, double *right_edge,
               bool *periodic, int64_t data_version)
        KDTree(double *pts, uint64_t *idx, uint64_t n, uint32_t m,
               uint32_t leafsize0, double *left_edge, double *right_edge,
               bool *periodic, int64_t data_version,
               bool use_sliding_midpoint)
        KDTree(istream &ist)
        void serialize(ostream &os)
        double* wrap_pos(double* pos) noexcept nogil
        vector[uint32_t] get_neighbor_ids(double* pos) noexcept nogil
        Node* search(double* pos) noexcept nogil
        Node* search(double* pos, bool dont_wrap) noexcept nogil
        void consolidate_edges(double *leaves_le, double *leaves_re)


cdef class PyNode:
    cdef Node *_node
    cdef readonly np.uint32_t id
    cdef readonly np.uint64_t npts
    cdef readonly np.uint32_t ndim
    cdef readonly np.uint32_t num_leaves
    cdef readonly np.uint64_t start_idx
    cdef readonly np.uint64_t stop_idx
    cdef double *_domain_width
    cdef readonly object left_neighbors, right_neighbors
    cdef void _init_node(self, Node* node, uint32_t num_leaves,
                         double *domain_width)

cdef class PyKDTree:
    cdef KDTree *_tree
    cdef readonly uint64_t npts
    cdef readonly uint32_t ndim
    cdef readonly uint32_t num_leaves
    cdef readonly uint32_t leafsize
    cdef readonly int64_t data_version
    cdef double *_left_edge
    cdef double *_right_edge
    cdef bool *_periodic
    cdef readonly object leaves
    cdef readonly object _idx
    cdef void _init_tree(self, KDTree* tree)
    cdef void _make_tree(self, double *pts, bool use_sliding_midpoint)
    cdef void _make_leaves(self)
    cdef np.ndarray[np.uint32_t, ndim=1] _get_neighbor_ids(self, np.ndarray[double, ndim=1] pos)
    cdef np.ndarray[np.uint32_t, ndim=1] _get_neighbor_ids_3(self, np.float64_t pos[3])
    cdef PyNode _get(self, np.ndarray[double, ndim=1] pos)



"""
AMRVAC-specific fields

"""

import functools

import numpy as np

from yt.fields.field_info_container import FieldInfoContainer
from yt.units import dimensions
from yt.utilities.logger import ytLogger as mylog

# We need to specify which fields we might have in our dataset.  The field info
# container subclass here will define which fields it knows about.  There are
# optionally methods on it that get called which can be subclassed.

direction_aliases = {
    "cartesian": ("x", "y", "z"),
    "polar": ("r", "theta", "z"),
    "cylindrical": ("r", "z", "theta"),
    "spherical": ("r", "theta", "phi"),
}


def _velocity(field, data, idir, prefix=None):
    """Velocity = linear momentum / density"""
    # This is meant to be used with functools.partial to produce
    # functions with only 2 arguments (field, data)
    # idir : int
    #    the direction index (1, 2 or 3)
    # prefix : str
    #    used to generalize to dust fields
    if prefix is None:
        prefix = ""
    moment = data["gas", "%smoment_%d" % (prefix, idir)]
    rho = data["gas", f"{prefix}density"]

    mask1 = rho == 0
    if mask1.any():
        mylog.info(
            "zeros found in %sdensity, "
            "patching them to compute corresponding velocity field.",
            prefix,
        )
        mask2 = moment == 0
        if not ((mask1 & mask2) == mask1).all():
            raise RuntimeError
        rho[mask1] = 1
    return moment / rho


code_density = "code_mass / code_length**3"
code_moment = "code_mass / code_length**2 / code_time"
code_pressure = "code_mass / code_length / code_time**2"


class AMRVACFieldInfo(FieldInfoContainer):
    # for now, define a finite family of dust fields (up to 100 species)
    MAXN_DUST_SPECIES = 100
    known_dust_fields = [
        ("rhod%d" % idust, (code_density, ["dust%d_density" % idust], None))
        for idust in range(1, MAXN_DUST_SPECIES + 1)
    ] + [
        (
            "m%dd%d" % (idir, idust),
            (code_moment, ["dust%d_moment_%d" % (idust, idir)], None),
        )
        for idust in range(1, MAXN_DUST_SPECIES + 1)
        for idir in (1, 2, 3)
    ]
    # format: (native(?) field, (units, [aliases], display_name))
    # note: aliases will correspond to "gas" typed fields
    # whereas the native ones are "amrvac" typed
    known_other_fields = (
        ("rho", (code_density, ["density"], None)),
        ("m1", (code_moment, ["moment_1"], None)),
        ("m2", (code_moment, ["moment_2"], None)),
        ("m3", (code_moment, ["moment_3"], None)),
        ("e", (code_pressure, ["energy_density"], None)),
        ("b1", ("code_magnetic", ["magnetic_1"], None)),
        ("b2", ("code_magnetic", ["magnetic_2"], None)),
        ("b3", ("code_magnetic", ["magnetic_3"], None)),
        ("Te", ("code_temperature", ["temperature"], None)),
        *known_dust_fields,
    )

    known_particle_fields = ()

    def _setup_velocity_fields(self, idust=None):
        if idust is None:
            dust_flag = dust_label = ""
        else:
            dust_flag = "d%d" % idust
            dust_label = "dust%d_" % idust

        us = self.ds.unit_system
        for idir, alias in enumerate(direction_aliases[self.ds.geometry], start=1):
            if ("amrvac", "m%d%s" % (idir, dust_flag)) not in self.field_list:
                break
            velocity_fn = functools.partial(_velocity, idir=idir, prefix=dust_label)
            self.add_field(
                ("gas", f"{dust_label}velocity_{alias}"),
                function=velocity_fn,
                units=us["velocity"],
                dimensions=dimensions.velocity,
                sampling_type="cell",
            )
            self.alias(
                ("gas", "%svelocity_%d" % (dust_label, idir)),
                ("gas", f"{dust_label}velocity_{alias}"),
                units=us["velocity"],
            )
            self.alias(
                ("gas", f"{dust_label}moment_{alias}"),
                ("gas", "%smoment_%d" % (dust_label, idir)),
                units=us["density"] * us["velocity"],
            )

    def _setup_dust_fields(self):
        idust = 1
        imax = self.__class__.MAXN_DUST_SPECIES
        while ("amrvac", "rhod%d" % idust) in self.field_list:
            if idust > imax:
                mylog.error(
                    "Only the first %d dust species are currently read by yt. "
                    "If you read this, please consider issuing a ticket. ",
                    imax,
                )
                break
            self._setup_velocity_fields(idust)
            idust += 1
        n_dust_found = idust - 1

        us = self.ds.unit_system
        if n_dust_found > 0:

            def _total_dust_density(field, data):
                tot = np.zeros_like(data["gas", "density"])
                for idust in range(1, n_dust_found + 1):
                    tot += data["dust%d_density" % idust]
                return tot

            self.add_field(
                ("gas", "total_dust_density"),
                function=_total_dust_density,
                dimensions=dimensions.density,
                units=us["density"],
                sampling_type="cell",
            )

            def dust_to_gas_ratio(field, data):
                return data["gas", "total_dust_density"] / data["gas", "density"]

            self.add_field(
                ("gas", "dust_to_gas_ratio"),
                function=dust_to_gas_ratio,
                dimensions=dimensions.dimensionless,
                sampling_type="cell",
            )

    def setup_fluid_fields(self):
        from yt.fields.magnetic_field import setup_magnetic_field_aliases

        setup_magnetic_field_aliases(self, "amrvac", [f"mag{ax}" for ax in "xyz"])
        self._setup_velocity_fields()  # gas velocities
        self._setup_dust_fields()  # dust derived fields (including velocities)

        # fields with nested dependencies are defined thereafter
        # by increasing level of complexity
        us = self.ds.unit_system

        def _kinetic_energy_density(field, data):
            # devnote : have a look at issue 1301
            return 0.5 * data["gas", "density"] * data["gas", "velocity_magnitude"] ** 2

        self.add_field(
            ("gas", "kinetic_energy_density"),
            function=_kinetic_energy_density,
            units=us["density"] * us["velocity"] ** 2,
            dimensions=dimensions.density * dimensions.velocity**2,
            sampling_type="cell",
        )

        # magnetic energy density
        if ("amrvac", "b1") in self.field_list:

            def _magnetic_energy_density(field, data):
                emag = 0.5 * data["gas", "magnetic_1"] ** 2
                for idim in "23":
                    if ("amrvac", f"b{idim}") not in self.field_list:
                        break
                    emag += 0.5 * data["gas", f"magnetic_{idim}"] ** 2
                # in AMRVAC the magnetic field is defined in units where mu0 = 1,
                # such that
                # Emag = 0.5*B**2 instead of Emag = 0.5*B**2 / mu0
                # To correctly transform the dimensionality from gauss**2 -> rho*v**2,
                # we have to take mu0 into account. If we divide here, units when adding
                # the field should be us["density"]*us["velocity"]**2.
                # If not, they should be us["magnetic_field"]**2 and division should
                # happen elsewhere.
                emag /= 4 * np.pi
                # divided by mu0 = 4pi in cgs,
                # yt handles 'mks' and 'code' unit systems internally.
                return emag

            self.add_field(
                ("gas", "magnetic_energy_density"),
                function=_magnetic_energy_density,
                units=us["density"] * us["velocity"] ** 2,
                dimensions=dimensions.density * dimensions.velocity**2,
                sampling_type="cell",
            )

        # Adding the thermal pressure field.
        # In AMRVAC we have multiple physics possibilities:
        # - if HD/MHD + energy equation P = (gamma-1)*(e - ekin (- emag)) for (M)HD
        # - if HD/MHD but solve_internal_e is true in parfile, P = (gamma-1)*e for both
        # - if (m)hd_energy is false in parfile (isothermal), P = c_adiab * rho**gamma

        def _full_thermal_pressure_HD(field, data):
            # energy density and pressure are actually expressed in the same unit
            pthermal = (data.ds.gamma - 1) * (
                data["gas", "energy_density"] - data["gas", "kinetic_energy_density"]
            )
            return pthermal

        def _full_thermal_pressure_MHD(field, data):
            pthermal = (
                _full_thermal_pressure_HD(field, data)
                - (data.ds.gamma - 1) * data["gas", "magnetic_energy_density"]
            )
            return pthermal

        def _polytropic_thermal_pressure(field, data):
            return (data.ds.gamma - 1) * data["gas", "energy_density"]

        def _adiabatic_thermal_pressure(field, data):
            return data.ds._c_adiab * data["gas", "density"] ** data.ds.gamma

        pressure_recipe = None
        if ("amrvac", "e") in self.field_list:
            if self.ds._e_is_internal:
                pressure_recipe = _polytropic_thermal_pressure
                mylog.info("Using polytropic EoS for thermal pressure.")
            elif ("amrvac", "b1") in self.field_list:
                pressure_recipe = _full_thermal_pressure_MHD
                mylog.info("Using full MHD energy for thermal pressure.")
            else:
                pressure_recipe = _full_thermal_pressure_HD
                mylog.info("Using full HD energy for thermal pressure.")
        elif self.ds._c_adiab is not None:
            pressure_recipe = _adiabatic_thermal_pressure
            mylog.info("Using adiabatic EoS for thermal pressure (isothermal).")
            mylog.warning(
                "If you used usr_set_pthermal you should "
                "redefine the thermal_pressure field."
            )

        if pressure_recipe is not None:
            self.add_field(
                ("gas", "thermal_pressure"),
                function=pressure_recipe,
                units=us["density"] * us["velocity"] ** 2,
                dimensions=dimensions.density * dimensions.velocity**2,
                sampling_type="cell",
            )

            # sound speed and temperature depend on thermal pressure
            def _sound_speed(field, data):
                return np.sqrt(
                    data.ds.gamma
                    * data["gas", "thermal_pressure"]
                    / data["gas", "density"]
                )

            self.add_field(
                ("gas", "sound_speed"),
                function=_sound_speed,
                units=us["velocity"],
                dimensions=dimensions.velocity,
                sampling_type="cell",
            )
        else:
            mylog.warning(
                "e not found and no parfile passed, can not set thermal_pressure."
            )



import tempfile
from pathlib import Path

import pytest

from yt.data_objects.static_output import Dataset
from yt.data_objects.time_series import DatasetSeries
from yt.utilities.exceptions import YTUnidentifiedDataType
from yt.utilities.object_registries import output_type_registry


def test_pattern_expansion():
    file_list = [f"fake_data_file_{str(i).zfill(4)}" for i in range(10)]

    with tempfile.TemporaryDirectory() as tmpdir:
        tmp_path = Path(tmpdir)
        for file in file_list:
            (tmp_path / file).touch()

        pattern = tmp_path / "fake_data_file_*"
        expected = [str(tmp_path / file) for file in file_list]
        found = DatasetSeries._get_filenames_from_glob_pattern(pattern)
        assert found == expected

        found2 = DatasetSeries._get_filenames_from_glob_pattern(Path(pattern))
        assert found2 == expected


def test_no_match_pattern():
    with tempfile.TemporaryDirectory() as tmpdir:
        pattern = Path(tmpdir).joinpath("fake_data_file_*")
        with pytest.raises(FileNotFoundError):
            DatasetSeries._get_filenames_from_glob_pattern(pattern)


@pytest.fixture
def FakeDataset():
    class _FakeDataset(Dataset):
        """A minimal loadable fake dataset subclass"""

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        @classmethod
        def _is_valid(cls, *args, **kwargs):
            return True

        def _parse_parameter_file(self):
            return

        def _set_code_unit_attributes(self):
            return

        def set_code_units(self):
            i = int(Path(self.filename).name.split("_")[-1])
            self.current_time = i
            self.current_redshift = 1 / (i + 1)
            return

        def _hash(self):
            return

        def _setup_classes(self):
            return

    try:
        yield _FakeDataset
    finally:
        output_type_registry.pop("_FakeDataset")


@pytest.fixture
def fake_datasets():
    file_list = [f"fake_data_file_{i:04d}" for i in range(10)]
    with tempfile.TemporaryDirectory() as tmpdir:
        pfile_list = [Path(tmpdir) / file for file in file_list]
        sfile_list = [str(file) for file in pfile_list]
        for file in pfile_list:
            file.touch()
        pattern = Path(tmpdir) / "fake_data_file_*"

        yield file_list, pfile_list, sfile_list, pattern


def test_init_fake_dataseries(fake_datasets):
    file_list, pfile_list, sfile_list, pattern = fake_datasets

    # init from str pattern
    ts = DatasetSeries(pattern)
    assert ts._pre_outputs == sfile_list

    # init from Path pattern
    ppattern = Path(pattern)
    ts = DatasetSeries(ppattern)
    assert ts._pre_outputs == sfile_list

    # init form str list
    ts = DatasetSeries(sfile_list)
    assert ts._pre_outputs == sfile_list

    # init form Path list
    ts = DatasetSeries(pfile_list)
    assert ts._pre_outputs == pfile_list

    # rejected input type (str repr of a list) "[file1, file2, ...]"
    with pytest.raises(FileNotFoundError):
        DatasetSeries(str(file_list))

    # finally, check that ts[0] fails to actually load
    with pytest.raises(YTUnidentifiedDataType):
        ts[0]


def test_init_fake_dataseries2(FakeDataset, fake_datasets):
    _file_list, _pfile_list, _sfile_list, pattern = fake_datasets
    ds = DatasetSeries(pattern)[0]
    assert isinstance(ds, FakeDataset)

    ts = DatasetSeries(pattern, my_unsupported_kwarg=None)

    with pytest.raises(TypeError):
        ts[0]


def test_get_by_key(FakeDataset, fake_datasets):
    _file_list, _pfile_list, sfile_list, pattern = fake_datasets
    ts = DatasetSeries(pattern)

    Ntot = len(sfile_list)

    t = ts[0].quan(1, "code_time")

    assert sfile_list[0] == ts.get_by_time(-t).filename
    assert sfile_list[0] == ts.get_by_time(t - t).filename
    assert sfile_list[1] == ts.get_by_time((0.8, "code_time")).filename
    assert sfile_list[1] == ts.get_by_time((1.2, "code_time")).filename
    assert sfile_list[Ntot - 1] == ts.get_by_time(t * (Ntot - 1)).filename
    assert sfile_list[Ntot - 1] == ts.get_by_time(t * Ntot).filename

    with pytest.raises(ValueError):
        ts.get_by_time(-2 * t, tolerance=0.1 * t)
    with pytest.raises(ValueError):
        ts.get_by_time(1000 * t, tolerance=0.1 * t)

    assert sfile_list[1] == ts.get_by_redshift(1 / 2.2).filename
    assert sfile_list[1] == ts.get_by_redshift(1 / 2).filename
    assert sfile_list[1] == ts.get_by_redshift(1 / 1.6).filename

    with pytest.raises(ValueError):
        ts.get_by_redshift(1000, tolerance=0.1)

    zmid = (ts[0].current_redshift + ts[1].current_redshift) / 2

    assert sfile_list[1] == ts.get_by_redshift(zmid, prefer="smaller").filename
    assert sfile_list[0] == ts.get_by_redshift(zmid, prefer="larger").filename
